﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Globalization;

//*************************************************************************
//*  Risultato Double, operazioni ammesse + - * / \                       *
//*  funzioni trigonometriche Sin(), Cos(), Tan(), Asin(), Acos(), Atan() *   
//*  Elevazione potenza ^, Sqrt(), Log10() Log(), parentesi ()            *
//*  Nella classe metodi per Tokenizzare e valutare i meno unari          *
//*************************************************************************

//********************************** METODI ESPOSTI ************************************
//* ClassParser  
//* Tokenizza [String] (tokenizza l'espressione per la valutazione da ParseEspressione)
//* DeTokenizza [String] (detokenizza l'espressione)
//* AggiustaSegni [String] (valuta segni accoppiati, non serve al di fuori della classe)
//* ParseEspressione [String] (valuta un'espressione matematica tokenizzata)
//************************************************************************************** 

public class ClassParser
{
    public string Tokenizza(string Espressione)
    {
        Espressione = Espressione.Replace(" ", ""); // rimuovo eventuali spazi
        if (CultureInfo.CurrentCulture.Name.ToLower() == "en-us")
        {
            Espressione = Espressione.Replace(",", "."); //punti decimali
            CulturaInfo = "Parser CB 2020   " + CultureInfo.CurrentCulture.Name + "  decimal commas are converted in decimal dots";
        }
        else
        {
            Espressione = Espressione.Replace(".", ","); //virgole decimali
            CulturaInfo = "Parser CB 2020   " + CultureInfo.CurrentCulture.Name + "  eventuali punti sono stati convertiti in virgole decimali";
        }
        string InputToken = Espressione.ToLower().Replace("asin", Token.Asin).Replace("acos", Token.Acos).Replace("atan", Token.Atan).Replace("sin", Token.Sin).Replace("cos", Token.Cos).Replace("tan", Token.Tan).Replace("log10", Token.log10).Replace("log", Token.log).Replace("sqrt", Token.Sqrt);
        return MenoUnari(InputToken); // la stringa restituita può essere valutata da ParseEspressione
    }
    public string DeTokenizza(string Espressione)
    {
        return Espressione.ToLower().Replace(Token.Asin, "asin").Replace(Token.Acos, "acos").Replace(Token.Atan, "atan").Replace(Token.Sin, "sin").Replace(Token.Cos, "cos").Replace(Token.Tan, "tan").Replace(Token.log10, "log10").Replace(Token.log, "log").Replace(Token.Sqrt, "sqrt");
    }

    public string AggiustaSegni(string Espressione)
    {
        return Espressione.Replace("+-", "-").Replace("--", "+").Replace("*+", "*").Replace("/+", "/").Replace("++", "+").Replace("-+", "-"); // Cambia i segni se duplicati
    }

    public double ParseEspressione(string Espressione)
    {
        double tempParseEspressione = 0;

        Espressione = AggiustaSegni(Espressione); // cambia i segni che si duplicano

        // operazioni con rispetto precedenza
        // parentesi
        int pos = Espressione.IndexOf(Token.p_chiusa);
        if (pos > -1) // Passa ricorsivamente quanto contenuto nelle parentesi
        {
            Espressione = Parentesi(Espressione, pos);
        }

        // potenze
        // Risolve le potenze da sinistra a destra non ricorsivo
        while (Espressione.Contains(Token.potenza) == true)
        {
            for (int Operatore = 0; Operatore < Espressione.Length; Operatore++)
            {
                if (Espressione[Operatore] == Token.potenza[0])
                {
                    Espressione = Operazioni(Espressione, Operatore, Token.potenza);
                    break;
                }
            }
        }

        // moltiplicazione e divisioni in sequenza da sinistra a destra non ricorsivo
        while (Espressione.Contains(Token.diviso) == true | Espressione.Contains(Token.per))
        {
            for (int Operatore = 0; Operatore < Espressione.Length; Operatore++)
            {
                if (Espressione[Operatore] == Token.diviso[0])
                {
                    Espressione = Operazioni(Espressione, Operatore, Token.diviso);
                    break;
                }
                if (Espressione[Operatore] == Token.per[0])
                {
                    Espressione = Operazioni(Espressione, Operatore, Token.per);
                    break;
                }
            }
        }

        // divisione intera
        // divisioni intere da sinistra a destra non ricorsivo
        while (Espressione.Contains(Token.divintera))
        {
            for (int Operatore = 0; Operatore < Espressione.Length; Operatore++)
            {
                if (Espressione[Operatore] == Token.divintera[0])
                {
                    Espressione = Operazioni(Espressione, Operatore, Token.divintera);
                    break;
                }
            }
        }

        // Somme e sottrazioni in sequenza da sinistra a destra non ricorsivo
        double contr = 0;
        while (Espressione.Contains(Token.meno) == true || Espressione.Contains(Token.piu) == true)
        {
            if (double.TryParse(Espressione, out contr) || Espressione == "+") // è rimaso solo un numero negativo o è stato digitato solo un +
            {
                break;
            }
            for (int Operatore = 1; Operatore < Espressione.Length; Operatore++) // da 1 perché - e + prima dei numeri sono ammessi
            {
                if (Espressione[Operatore] == Token.meno[0])
                {
                    Espressione = Operazioni(Espressione, Operatore, Token.meno);
                    break;
                }
                if (Espressione[Operatore] == Token.piu[0])
                {
                    Espressione = Operazioni(Espressione, Operatore, Token.piu);
                    break;
                }
            }
        }

        double ris = 0; // tutte le operazioni sono state risolte
        if (double.TryParse(Espressione, out ris)) // controllo che il risultato sia un double
        {
            tempParseEspressione = ris; // se è un double lo restituisco alla funzione
        }
        else
        {
            throw new Exception("Syntax error"); // altrimenti restituisco un errore
        }
        return tempParseEspressione;
    }

    //********************************* PROPRIETA' ESPOSTE ***********************************
    //* Decimali [String] es. "0.000000" (=6 decimali)
    //* CulturaInfo [string] (restituisce l'interpretazione dei . e , decimali)
    //* Progressione [List<string>] (progresso, inizializzare prima di usare ParseEspressione)
    //****************************************************************************************

    public string Decimali { get; set; }
    public string CulturaInfo { get; set; }
    public List<string> Progressione { get; set; }

    //********************************* METODI INTERNI ALLA CLASSE **********************************
    //* Parentesi [String] estrae le espressioni tra parantesi o risolve le funzioni trigonometriche
    //* Operazioni [string] risolve le operazioni matematiche
    //* ControlloSintassiDX [string] sintaddi a destra
    //* ControlloSintassiSX [string] sintaddi a sinistra
    //* MenoUnari [string] sostituisce i meno unari con -1*
    //***********************************************************************************************


    //**** Ordine di esecuzione 1
    private string Parentesi(string EspressioneP, int Dove)
    {
        string tempParentesi = null;
        short Caratteri = 0;
        short CurPos = 0;
        string StringSX = "";
        string TraDueParentesi = "";
        double dValore = 0;
        tempParentesi = EspressioneP;

        if (Dove > 0)
        {
            Caratteri = 0;
            CurPos = (short)(Dove + 1);
            do //cicla tra le parentesi
            {
                CurPos = (short)(CurPos - 1);
                if (CurPos == 0) // inizio
                {
                    break;
                }
                if (EspressioneP[CurPos] != Token.p_aperta[0])
                {
                    TraDueParentesi = EspressioneP.Substring(CurPos, Caratteri);
                }
                Caratteri = (short)(Caratteri + 1);
            } while (!(EspressioneP[CurPos - 1] == Token.p_aperta[0]));

            if (EspressioneP.Length - Dove > 1)
            {
                ControlloSintassiDX(EspressioneP, CurPos, Dove);
            }

            if (CurPos > 0)
            {
                if (double.TryParse(TraDueParentesi, out dValore))
                {
                    string tokens = "";
                    if (CurPos != 1)
                    {
                        StringSX = EspressioneP.Substring(0, CurPos - 2);
                        tokens = EspressioneP[CurPos - 2].ToString();
                    }
                    switch (tokens)
                    {
                        case Token.Asin:
                            Progressione.Add("Asin(" + TraDueParentesi + ") = " + Math.Asin(dValore).ToString());
                            ControlloSintassiSX(EspressioneP, CurPos, "asin");
                            tempParentesi = ParseEspressione(StringSX + Math.Asin(dValore).ToString(Decimali) + EspressioneP.Substring(Dove + 1)).ToString();
                            break;
                        case Token.Acos:
                            Progressione.Add("Acos(" + TraDueParentesi + ") = " + Math.Acos(dValore).ToString());
                            ControlloSintassiSX(EspressioneP, CurPos, "acos");
                            tempParentesi = ParseEspressione(StringSX + Math.Acos(dValore).ToString(Decimali) + EspressioneP.Substring(Dove + 1)).ToString();
                            break;
                        case Token.Atan:
                            Progressione.Add("Atan(" + TraDueParentesi + ") = " + Math.Atan(dValore).ToString());
                            ControlloSintassiSX(EspressioneP, CurPos, "atan");
                            tempParentesi = ParseEspressione(StringSX + Math.Atan(dValore).ToString(Decimali) + EspressioneP.Substring(Dove + 1)).ToString();
                            break;
                        case Token.Sin:
                            Progressione.Add("Sin(" + TraDueParentesi + ") = " + Math.Sin(dValore).ToString());
                            ControlloSintassiSX(EspressioneP, CurPos, "sin");
                            tempParentesi = ParseEspressione(StringSX + Math.Sin(dValore).ToString(Decimali) + EspressioneP.Substring(Dove + 1)).ToString();
                            break;
                        case Token.Cos:
                            Progressione.Add("Cos(" + TraDueParentesi + ") = " + Math.Cos(dValore).ToString());
                            ControlloSintassiSX(EspressioneP, CurPos, "cos");
                            tempParentesi = ParseEspressione(StringSX + Math.Cos(dValore).ToString(Decimali) + EspressioneP.Substring(Dove + 1)).ToString();
                            break;
                        case Token.Tan:
                            Progressione.Add("Tan(" + TraDueParentesi + ") = " + Math.Tan(dValore).ToString());
                            ControlloSintassiSX(EspressioneP, CurPos, "tan");
                            tempParentesi = ParseEspressione(StringSX + Math.Tan(dValore).ToString(Decimali) + EspressioneP.Substring(Dove + 1)).ToString();
                            break;
                        case Token.log10:
                            Progressione.Add("Log10(" + TraDueParentesi + ") = " + Math.Log10(dValore).ToString());
                            ControlloSintassiSX(EspressioneP, CurPos, "log10");
                            tempParentesi = ParseEspressione(StringSX + Math.Log10(dValore).ToString(Decimali) + EspressioneP.Substring(Dove + 1)).ToString();
                            break;
                        case Token.Sqrt:
                            Progressione.Add("Sqrt(" + TraDueParentesi + ") = " + Math.Sqrt(dValore).ToString());
                            ControlloSintassiSX(EspressioneP, CurPos, "sqrt");
                            tempParentesi = ParseEspressione(StringSX + Math.Sqrt(dValore).ToString(Decimali) + EspressioneP.Substring(Dove + 1)).ToString();
                            break;
                        case Token.log:
                            Progressione.Add("Log(" + TraDueParentesi + ") = " + Math.Log(dValore).ToString());
                            ControlloSintassiSX(EspressioneP, CurPos, "log");
                            tempParentesi = ParseEspressione(StringSX + Math.Log(dValore).ToString(Decimali) + EspressioneP.Substring(Dove + 1)).ToString();
                            break;
                        case Token.p_aperta: // parentesi multiple
                            Progressione.Add("(" + TraDueParentesi + ") = " + (dValore).ToString());
                            tempParentesi = ParseEspressione(StringSX + "(" + TraDueParentesi + EspressioneP.Substring(Dove + 1)).ToString();
                            break;
                        default:
                            ControlloSintassiSX(EspressioneP, CurPos + 1, tokens);
                            tempParentesi = ParseEspressione(EspressioneP.Substring(0, CurPos - 1) + TraDueParentesi + EspressioneP.Substring(Dove + 1)).ToString();
                            break;
                    }
                }
                else
                {
                    int r = 0; // r cambia la posizione di concatenamento in base al token trovato
                    if (CurPos > 1)
                    {
                        string tokens = EspressioneP[CurPos - 2].ToString(); // prima della parentesi aperta
                        Progressione.Add("Token " + tokens);
                        if (tokens == Token.meno || tokens == Token.piu || tokens == Token.per || tokens == Token.diviso || tokens == Token.divintera || tokens == Token.potenza || tokens == Token.p_aperta)
                        {
                            r = 1;
                        }
                        else if (tokens == Token.Acos || tokens == Token.Asin || tokens == Token.Atan || tokens == Token.Cos || tokens == Token.log10 || tokens == Token.Sin || tokens == Token.Sqrt || tokens == Token.Tan || tokens == Token.log)
                        {
                            r = 0;
                        }
                        else
                        {
                            Progressione.Add("Syntax error " + tokens + "(");
                            throw new Exception("Syntax error ;" + tokens + "("); // non c'è un simbolo valido a Sx
                        }
                    }
                    else
                    {
                        r = 1;
                    }
                    StringSX = EspressioneP.Substring(0, CurPos - r);
                    EspressioneP = StringSX + ParseEspressione(TraDueParentesi).ToString(Decimali) + EspressioneP.Substring(Dove + r); // ricorsione ParseEspressione
                    int pos = EspressioneP.IndexOf(Token.p_chiusa);
                    if (pos > -1)
                    {
                        tempParentesi = Parentesi(EspressioneP, pos); // Passa ricorsivamente quanto contenuto nelle parentesi, ' ricorsione Parentesi
                    }
                    else
                    {
                        tempParentesi = ParseEspressione(EspressioneP).ToString(); // non ci sono parentesi, si risolvono gli operatori matematici, ' ricorsione ParseEspressione
                    }
                }
            }
        }
        return tempParentesi;
    }
    //**** Ordine di esecuzione 2
    private string Operazioni(string Espressione, int Dove, string tipo)
    {
        string tempOperazioni = null;
        if (Dove == 0)
        {
            Progressione.Add("Syntax error " + Espressione.Substring(0, 2));
            throw new Exception("Syntax error ;" + Espressione.Substring(0, 2)); // non c'è un simbolo valido
        }

        // Logica
        // l'espressione a questo punto può contenere solo + - * / ^, perché le altre operazioni sono state svolte.
        // Per evitare di scorrere continuamente avanti ed indietro l'espressione, per isolare i numeri, con
        // svariati controlli, vedi numeri negativi e inizio/fine espressione, eseguo una scansione completa dell'espressione,
        // memorizzando in "Separazioni" tutte le posizioni dei simboli matematici rimasti
        // tra le posizioni dei simboli matematici ci sono sicuramente i numeri da calcolare.
        // per eseguire l'operazione giusta serve anche sapere l'indice in "Separazioni" di dove è memorizzata la posizione
        // del simbolo corrente, l'indice viene messo in "PosList", e in quella posizione c'è "Dove"

        List<int> Separazioni = new List<int>(); // lista che contiene l'indice base zero di tutte le separazioni
        Separazioni.Add(-1); // default, è la prima separazione
        int PosList = 0; // posizione dell'operazione in corso nella lista delle separazioni
        int PosEspr = 0; // ricerca delle separazioni
        int Indice = 0; // scandisce la lista "Separazioni"
        for (PosEspr = 1; PosEspr < Espressione.Length; PosEspr++) // salto la posizione zero perché ci potrebbe essere un numero ma anche un segno meno
        {
            string op = Espressione[PosEspr].ToString();
            if (op == Token.piu || op == Token.meno || op == Token.per || op == Token.diviso || op == Token.divintera || op == Token.potenza)
            {
                Indice += 1;
                Separazioni.Add(PosEspr); // è uno dei simboli riconosciuti, ne memorizzo la posizione
                if (PosEspr == Dove) // tengo traccia della posizione in "PosList"
                {
                    PosList = Indice;
                }
                PosEspr += 1; // salto cifra o segno meno
            }
        }
        Separazioni.Add(PosEspr); // default, è l'ultima separazione, la lunghezza della stringa

        // con tutte le posizioni memorizzate, fare i calcoli tra i numeri giusti diventa semplice
        // risolve 5 operazioni matematiche + - * / ^
        tempOperazioni = Espressione; // all'inizio risposta con la striga intera
        string StringSX = Espressione.Substring(0, Separazioni[PosList - 1] + 1);
        double NumSx = Convert.ToDouble(Espressione.Substring(Separazioni[PosList - 1] + 1, Dove - Separazioni[PosList - 1] - 1));
        string StringDX = Espressione.Substring(Separazioni[PosList + 1]);
        double NumDx = Convert.ToDouble(Espressione.Substring(Dove + 1, Separazioni[PosList + 1] - Dove - 1));
        switch (tipo)
        {
            case Token.potenza:
                Progressione.Add("Pow(" + NumSx + "^" + NumDx + ") = " + Math.Pow(NumSx, NumDx).ToString());
                tempOperazioni = StringSX + Math.Pow(NumSx, NumDx).ToString(Decimali) + StringDX;

                break;
            case Token.per:
                Progressione.Add("(" + NumSx + "*" + NumDx + ") = " + (NumSx * NumDx).ToString());
                tempOperazioni = StringSX + (NumSx * NumDx).ToString(Decimali) + StringDX;

                break;
            case Token.diviso:
                Progressione.Add("(" + NumSx + "/" + NumDx + ") = " + (NumSx / NumDx).ToString());
                tempOperazioni = StringSX + (NumSx / NumDx).ToString(Decimali) + StringDX;

                break;
            case Token.divintera:
                Progressione.Add("(" + NumSx + "\\" + NumDx + ") = " + (Convert.ToInt32(NumSx) / Convert.ToInt32(NumDx)).ToString());
                tempOperazioni = StringSX + (Convert.ToInt32(NumSx) / Convert.ToInt32(NumDx)).ToString(Decimali) + StringDX;

                break;
            case Token.piu:
                Progressione.Add("(" + NumSx + "+" + NumDx + ") = " + (NumSx + NumDx).ToString());
                tempOperazioni = StringSX + (NumSx + NumDx).ToString(Decimali) + StringDX;

                break;
            case Token.meno:
                Progressione.Add("(" + NumSx + "-" + NumDx + ") = " + (NumSx - NumDx).ToString());
                tempOperazioni = StringSX + (NumSx - NumDx).ToString(Decimali) + StringDX;

                break;
        }
        return AggiustaSegni(tempOperazioni);
    }

    private void ControlloSintassiDX(string EspressioneS, int curpos, int dove)
    {
        // controllo a destra di tutte le parentesi
        string tokens = EspressioneS[dove + 1].ToString(); // controllo dopo della parentesi chiusa
        if (tokens != Token.meno & tokens != Token.piu & tokens != Token.per & tokens != Token.diviso & tokens != Token.divintera & tokens != Token.potenza & tokens != Token.p_chiusa)
        {
            Progressione.Add("Syntax error " + ")" + tokens);
            throw new Exception("Syntax error ;" + ")" + tokens); // non c'è un simbolo valido
        }
    }

    private void ControlloSintassiSX(string EspressioneS, int curpos, string tipo)
    {
        // controllo a sinistra di tutte le funzioni
        if (curpos > 2)
        {
            string tokens = EspressioneS[curpos - 3].ToString(); // controllo prima della funzione
            if (tokens != Token.meno & tokens != Token.piu & tokens != Token.per & tokens != Token.diviso & tokens != Token.divintera & tokens != Token.potenza & tokens != Token.p_aperta)
            {
                Progressione.Add("Syntax error " + tokens + tipo);
                throw new Exception("Syntax error ;" + tokens + tipo); // non c'è un simbolo valido
            }
        }
    }

    private string MenoUnari(string Espressione)
    {
        //sostituisce i meno unari, - unario = -1*
        StringBuilder sbUnario = new StringBuilder();
        int Paperta = 0;
        int Pchiusa = 0;
        for (int unario = 0; unario < Espressione.Length; unario++)
        {
            if (Espressione[unario] == '^')
            {
                sbUnario.Append(Espressione[unario]);
                unario += 1;
                if (Espressione[unario] == '-') // è un esponente negativo, non è un meno unario
                {
                    sbUnario.Append(Espressione[unario]);
                    unario += 1;
                }
            }
            if (Espressione[unario] == '-') // è un meno unario
            {
                sbUnario.Append("-1*");
            }
            else
            {
                sbUnario.Append(Espressione[unario]);
            }
            // controllo bilanciamento parentesi
            if (Espressione[unario] == '(') { Paperta++; }
            if (Espressione[unario] == ')') { Pchiusa++; }
        }
        if (Paperta != Pchiusa)
        {
            if (Paperta > Pchiusa)
            {
                Progressione.Add("parentesi sbilanciate (");
                throw new Exception("Parentesi sbilanciate ;("); // più parentesi aperte che chiuse
            }
            else
            {
                Progressione.Add("parentesi sbilanciate )");
                throw new Exception("Parentesi sbilanciate ;)"); // più parentesi chiuse che aperte
            }
         }
        return sbUnario.ToString();
    }

    // Tokens usabili per implementare altri operatori ª«¬­®¯°±²³´µ¶

    private struct Token
    {
        public const string piu = "+";
        public const string meno = "-";
        public const string diviso = "/";
        public const string divintera = "\\";
        public const string per = "*";
        public const string potenza = "^";
        public const string p_aperta = "(";
        public const string p_chiusa = ")";
        public const string Asin = "¡";
        public const string Acos = "¢";
        public const string Atan = "£";
        public const string Sin = "¤";
        public const string Cos = "¥";
        public const string Tan = "¦";
        public const string log10 = "§";
        public const string Sqrt = "¨";
        public const string log = "©";
    }
}
