﻿using System;
using System.CodeDom.Compiler;
using System.Reflection;
using System.Windows.Forms;

//***********************************************************************************************
//*                                     Carlo Barabucci 2020                                    *
//* Esempio d'uso di ClassParser che valuta un'espressione matematica in una stringa.           *
//* Il risultato viene confrontato con il valutatore implementato in .Net, sicuro nel risultato *
//* ma lento nella risposta, perché compilato ad ogni interrogazione in ram                     *  
//***********************************************************************************************

namespace ParserEspressioneMatematica
{
    internal partial class Form1 : Form
    {
        private ClassParser Parser = new ClassParser();

        //********** PROPRIETA' ClassParser
        //* Parser.Decimali [String] es. "0.000000" (6 decimali)
        //* Parser.Progressione [list(Of String)] (inizializzare con new prima di usare)
        //* Parser.CulturaInfo [string] (restituisce l'interpretazione dei . e , decimali)
        //********** METODI ClassParser
        //* Parser.Tokenizza [String] (tokenizza l'espressione per una corretta valutazione da ClassParser)
        //* Parser.DeTokenizza [String] (detokenizza l'espressione)
        //* Parser.ParseEspressione [String] (valuta un'espressione matematica tokenizzata)
        //* Parser.AggiustaSegni [String] (valuta segni accoppiati, non serve richiamarla)

        public Form1()
        {
            InitializeComponent();
            ButRISOLVI.Click += ButRISOLVI_Click;
            Button1.Click += Button1_Click;
            TexInCS.KeyDown += TexInCS_KeyDown;
            ListBox1.SelectedIndexChanged += ListBox1_SelectedIndexChanged;
            ComboBox1.SelectedIndexChanged += ComboBox1_SelectedIndexChanged;
            Parser.Decimali = "0.".PadRight(15 + 2, '0'); // 15 decimali default
        }

        private void ButRISOLVI_Click(System.Object eventSender, System.EventArgs eventArgs)
        {
            Label2.Text = "";
            this.Refresh();
            string Espressione = TexInCS.Text;

            //*** Richiesta a ClassParser
            try
            {
                Parser.Progressione = new System.Collections.Generic.List<string>(); // conterrà la progressione dei calcoli
                double Risultato = Parser.ParseEspressione(Parser.Tokenizza(Espressione));
                Label2.Text = Risultato.ToString();
                richTextBox1.Lines = Parser.Progressione.ToArray();
                this.Text = Parser.CulturaInfo;
                this.Refresh();
            }
            catch (Exception ex)
            {
                string[] dati = ex.Message.Split(';'); // divido l'errore ritornato
                if (dati.Length == 2) // controllo che ci siano due stringhe
                {
                    TexInCS.Focus(); 
                    string ResErr = Parser.DeTokenizza(dati[1]); // in dati[1] c'è la porzione di stringa dove è avvenuto l'errore
                    // cerco la porzione di striga ritornata e detokenizzata in TexInCS
                    int pos = TexInCS.Text.ToLower().IndexOf(ResErr);
                    if (pos > -1)
                    {   // seleziono il testo dove c'è l'errore
                        TexInCS.SelectionStart = pos;
                        TexInCS.SelectionLength = ResErr.Length;
                    }
                }
                Label2.Text = dati[0]; // in dati[0] c'è il tipo di errore di sistema o personale
                richTextBox1.Lines = Parser.Progressione.ToArray();
            }

            //**********************************
            //*** Richiesta a Framework .Net ***
            //**********************************
            Label3.Text = "";
            this.Refresh();
            // alcune modifiche per rendere l'input compatibile con il Framework
            string InputValidato = TexInCS.Text.ToLower().Replace(",", ".").Replace("asin", "Math.ASin").Replace("acos", "Math.ACos").Replace("atan", "Math.ATan").Replace("log10", "Math.Log10").Replace("tan", "Math.Tan").Replace("cos", "Math.Cos").Replace("sin", "Math.Sin").Replace("sqrt", "Math.Sqrt").Replace("log", "Math.Log").Replace("Logn", "Math");
            Label4.Text = InputValidato;
            try
            {
                Label3.Text = Convert.ToString(ValutazioneEspressioneNET(InputValidato)); // dal Framework .Net, per verifica
            }
            catch (Exception ex)
            {
                Label3.Text = ex.Message;
            }
        }

        private void TexInCS_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyValue == 13)
            {
                ButRISOLVI_Click(ButRISOLVI, e);
            }
        }

        private void ListBox1_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyValue == 13)
            {
                ButRISOLVI_Click(ButRISOLVI, e);
            }
        }

        private void ListBox1_SelectedIndexChanged(object sender, EventArgs e)
        {
            TexInCS.Text = ListBox1.SelectedItem.ToString();
            ButRISOLVI_Click(ButRISOLVI, e);
        }

        private void ComboBox1_SelectedIndexChanged(object sender, EventArgs e)
        {
            int quanti = ComboBox1.SelectedIndex;
            Parser.Decimali = "0.".PadRight(quanti + 2, '0');
            ButRISOLVI_Click(ButRISOLVI, e);
            ListBox1.Focus();
        }

        private void Button1_Click(object sender, EventArgs e)
        {
            richTextBox1.Visible = !richTextBox1.Visible;
        }

        public object ValutazioneEspressioneNET(string syntax)
        {
            //Creo il compilatore VB e relativi paramentri
            Microsoft.VisualBasic.VBCodeProvider provider = new Microsoft.VisualBasic.VBCodeProvider();
            CompilerParameters param = new CompilerParameters();
            param.GenerateInMemory = true;
            param.GenerateExecutable = false;
            param.ReferencedAssemblies.Add("System.dll");

            //Creo una classe temporanea
            string src = "Imports System" + Environment.NewLine +
            "Class Evaluator" + Environment.NewLine +
            "Public Function Eval() As Object" + Environment.NewLine +
            "Return " + syntax + Environment.NewLine +
            "End Function" + Environment.NewLine +
            "End Class";

            //Compilo dinamicamente la classe
            CompilerResults results = provider.CompileAssemblyFromSource(param, src);
            if (results.Errors.Count > 0)
            {
                throw new ArgumentException("Syntax error");
            }
            else
            {
                //Creo istanza della classe Evaluator e invoco metodo Eval()
                object o = results.CompiledAssembly.CreateInstance("Evaluator");
                MethodInfo method = o.GetType().GetMethod("Eval");
                return method.Invoke(o, null);
            }
        }
    }
}


